/* Copyright Coraid, Inc. 2007.  All Rights Reserved. */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/errno.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <time.h>
#include "dat.h"

#define dprint(...) if (!debug) ; else fprintf(stderr, __VA_ARGS__)

struct aoemask {
	uchar dst[6];
	uchar src[6];
	uchar etype[2];
	uchar verflags;
	uchar error;
	uchar major[2];
	uchar minor;
	uchar cmd;
	uchar tag[4];
	uchar rid;
	uchar mcmd;
	uchar merror;
	uchar dircnt;
};

struct maskdir {
	uchar res;
	uchar dcmd;
	uchar mac[6];
};

enum {
	// mcmds
	MCread= 0,
	MCedit,

	// merrors
	MEunspec= 1,
	MEbaddir,
	MEfull,

	// dcmds
	MDnone= 0,
	MDadd,
	MDdel,

	AOEMASKCMD= 0xf0,
	AOE_ETYPE= 0x88a2,
	AOEF_RSP= (1<<3),
	AOEF_ERR= (1<<2),

	WAITSECS= 2,
};

int waitsecs = WAITSECS;
uchar cmdbuf[1536];
uchar rspbuf[1536];
struct aoemask *cmd = (struct aoemask *) cmdbuf;
struct aoemask *rsp = (struct aoemask *) rspbuf;
ulong shelf, slot;
ulong ncmdbuf;
u_int32_t tag;
char *ether;
uchar etheraddr[6];
int eflag;
int lflag;
int sflag;

char *aoe_errs[] = {
	"unknown error",
	"unrecognized command code",
	"bad argument parameter",
	"device unavailable",
	"config string present",
	"unsupported version"
};

char *aoemask_errs[] = {
	"unknown",
	"unspecified error",
	"bad dcmd directive",
	"mask list full"
};

void timewait(int);

void
usage(void)
{
	fprintf(stderr, "usage: %s shelf slot interface [+mac ...] [-mac ...]\n", progname);
	exits("usage");
}

void
setdirs(int argc, char *argv[])
{
	struct maskdir *ad, *adi;
	char *p;

	if (argc + !eflag > 255) {
		fprintf(stderr, "too many mac directives\n");
		exits("usage");
	}
	ad = adi = (struct maskdir *) (cmd+1);
	if (!eflag) {	// set the first directive to identity
		ad->dcmd = MDadd;
		memmove(ad->mac, etheraddr, 6);
		ad++;
	}
	while (argc-- > 0) {
		p = *argv++;
		switch (*p) {
		case '+':
			ad->dcmd = MDadd;
			break;
		case '-':
			ad->dcmd = MDdel;
			break;
		default:
			fprintf(stderr, "bad directive %s\n", p);
			usage();
		}
		if (parseether(ad->mac, p+1) < 0) {
			fprintf(stderr, "parseether failure for directive %s\n", p);
			exits("badmac");
		}
		ad++;
	}
	cmd->dircnt = ad - adi;
	ncmdbuf += cmd->dircnt * sizeof *ad;
}

void
hnput16(uchar *p, u_int16_t s)
{
	p[0] = s>>8;
	p[1] = s;
}

void
hnput32(uchar *p, u_int32_t n)
{
	p[0] = n>>24;
	p[1] = n>>16;
	p[2] = n>>8;
	p[3] = n;
}

ushort
nhget16(uchar *p)
{
	u_int16_t s;

	s = *p++;
	s <<= 8;
	s |= *p;
	return s;
}

u_int32_t
nhget32(uchar *p)
{
	u_int32_t n;
	int i;

	for (n=0, i=0; i<4; i++) {
		n <<= 8;
		n |= p[i] & 0xff;
	}
	return n;
}

void
sethdr(void)
{
	memset(cmd->dst, 0xff, 6);
	hnput16(cmd->etype, AOE_ETYPE);
	cmd->verflags = 0x10;
	hnput16(cmd->major, shelf);
	cmd->minor = slot;
	cmd->cmd = AOEMASKCMD;
	srand(time(0));
	tag = rand();
	tag |= 1<<31;	// tag space with high bit set is reserved for userspace
	hnput32(cmd->tag, tag);
}

void
dumprsp(void)
{
	struct maskdir *ad, *e;
	char buf[sizeof *e * 3];

	ad = (struct maskdir *) (rsp+1);
	e = ad + rsp->dircnt;
	if (rsp->merror) {
		if (rsp->merror > nelem(aoemask_errs))
			rsp->merror = 0;
		fprintf(stderr, "%d.%d mask error: %s\n",
			nhget16(rsp->major), rsp->minor, aoemask_errs[rsp->merror]);
		if (rsp->merror == MEbaddir) {
			htoa(buf, (uchar *)e, sizeof *e);
			buf[sizeof *e * 2] = '\0';
			fprintf(stderr, "\tdirective: %s\n", buf);
		}
		exits("mask error");
	}
	if (cmd->mcmd == MCedit)
	if (lflag == 0)
		return;
	printf("%d.%d", nhget16(rsp->major), rsp->minor);

	buf[12] = '\0';
	for (; ad<e; ad++) {
		htoa(buf, ad->mac, 6);
		printf(" %s", buf);
	}
	printf("\n");
}

/* returns the number of command responses received */
int
domask(void)
{
	int n, rx = 0;

	netsend(cmdbuf, ncmdbuf);
	if (sflag)
		return 0;
	timewait(waitsecs);
loop:
	n = netrecv();
	if (n < 0 && errno == EINTR) {
		alarm(0);
		return rx;
	}
	while ((n=netget(rspbuf, sizeof rspbuf)) > 0) {
		if (n < 60) {
			dprint("n: %d\n", n);
			continue;
		}
		if (nhget16(rsp->etype) != AOE_ETYPE) {
			dprint("etype: %04ulx\n", nhget16(rsp->etype));
			continue;
		}
		if (rsp->cmd != AOEMASKCMD) {
			dprint("cmd: %d\n", rsp->cmd);
			continue;
		}
		if (shelf != 0xffff && shelf != nhget16(rsp->major)) {
			dprint("shelf: %d\n", nhget16(rsp->major));
			continue;
		}
		if (slot != 0xff && slot != rsp->minor) {
			dprint("slot: %d\n", rsp->minor);
			continue;
		}
		if ((rsp->verflags & AOEF_RSP) == 0) {
			dprint("verflags: %d\n", rsp->verflags);
			continue;
		}
		if (nhget32(rsp->tag) != tag) {
			dprint("tag: %d, %d\n", nhget32(rsp->tag), tag);
			continue;
		}
		if (rsp->verflags & AOEF_ERR) {
			if (rsp->error >= nelem(aoe_errs))
				rsp->error = 0;
			fprintf(stderr, "aoe error: %s", aoe_errs[rsp->error]);
			exits("aoe error");
		}
		rx++;
		dumprsp();
		if (shelf != 0xffff)
		if (slot != 0xff) {
			alarm(0);
			return rx;
		}
		timewait(waitsecs);
	}
	goto loop;
}

void
catch(int sig)
{
}

void
timewait(int secs)	/* arrange for a sig_alarm signal after `secs' seconds */
{
	struct sigaction sa;
	
	memset(&sa, 0, sizeof sa);
	sa.sa_handler = catch;
	sa.sa_flags = SA_RESETHAND;
	sigaction(SIGALRM, &sa, NULL);
	alarm(secs);
}

void
exits(char *s)
{
	if(!s || !*s)
		exit(0);
	exit(1);
}

int
main(int argc, char **argv)
{
	int ch;
	unsigned char smac[6];
	
	progname = *argv;
	while ((ch = getopt(argc, argv, "+dels:vw:?")) != -1) {
		switch (ch) {
		case 'd':
			debug = 1;
			break;
		case 'e':
			eflag++;
			break;
		case 'l':
			lflag++;
			break;
		case 's':
			sflag++;
			if (parseether(smac, optarg) < 0) {
				fprintf(stderr, "error parsing source address %s\n", optarg);
				exits("parseether");
			}
			break;
		case 'v':
			fprintf(stderr, "%s\n", VERSION);
			return 0;
		case 'w':
			waitsecs = atoi(optarg);
			if (waitsecs <= 0) {
				fprintf(stderr, "Invalid w value, ignoring.\n");
				waitsecs = WAITSECS;
			}
			break;
		case '?':
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;
	if (argc < 3)
		usage();
	if (debug)
		fprintf(stderr, "debug is on\n");

	shelf = strtol(*argv, 0, 0);
	argv++, argc--;
	slot = strtol(*argv, 0, 0);
	argv++, argc--;
	ether = *argv;
	argv++, argc--;

	if (netopen(ether, AOE_ETYPE) < 0) {
		fprintf(stderr, "%s: can't netopen %s\n", progname, *argv);
		exits("open");
	}
	netgetaddr(etheraddr);
	if (sflag)
		netsetaddr(smac);

	sethdr();
	ncmdbuf = sizeof *cmd;
	if (argc > 0) {
		setdirs(argc, argv);
		cmd->mcmd = MCedit;
	} else
		cmd->mcmd = MCread;
	return !(domask() || sflag);	// if sflag, no rsp is ok
}

